"""
.. _Input:

The :ref:`Input <Input>` class creates immutable instances that describe the inputs of a single turn of a conversation.
There are some attributes that may have aggregations of previous turns like the :code:`slot_tracker` or entire :code:`history`.

.. admonition:: Why do I see :ref:`Input <Input>` and :ref:`Output <Output>` as inputs to all :ref:`Plugins <AbstractPlugin>`?

    It is a common pattern for all the plugins to require both as arguments. Since this could be confusing nomenclature, :ref:`Input <Input>`
    and :ref:`Output <Output>` bear meaning and even separation for the SLU API, **not for** :ref:`Plugins <AbstractPlugin>`.


Updates
-------

While writing plugins, we would need to update the attributes of :ref:`Input <Input>`, the following **doesn't** work!

.. ipython::

    In [1]: from dialogy.base import Input
       ...: from dialogy.utils import make_unix_ts

    In [2]: # Check the attributes in the object logged below. 
       ...: input_x = Input(utterances="hello world", lang="hi", timezone="Asia/Kolkata")

    In [3]: input_x

Issues with Frozen Instance Update
##################################

Now if we try the following:

.. ipython::
    :okexcept:

    In [4]: input_x.utterances = "hello"

We can see re-assigning values to attributes isn't allowed. 

Updating a frozen instance
##########################

We have to create new instances, but we have some syntax for it: 

.. ipython::

    In [1]: input_x = Input(utterances="hello world", lang="hi", timezone="Asia/Kolkata")

    In [2]: input_y = Input.from_dict({"utterances": "hello"})

    In [3]: input_y

but by doing this, we lost the :code:`lang` and :code:`timezone` attributes to system defaults.

Reusing an existing instance
############################

We can re-use an existing instance to create a new.
This way, we don't have to write every existing property on a previous :ref:`Input <Input>`.

.. ipython::

    In [1]: input_x = Input(utterances="hello world", lang="hi", timezone="Asia/Kolkata")

    In [2]: input_y = Input.from_dict({"utterances": "hello"}, reference=input_x)

    In [3]: input_y

Serialization
-------------

If there is a need to represent an :ref:`Input<Input>` as a `dict` we can do the following:

.. ipython::

    In [4]: input_y.json()

"""
from __future__ import annotations

from typing import Any, Dict, List, Optional

import attr

from dialogy.types import Utterance
from dialogy.utils import is_unix_ts, normalize


@attr.frozen
class Input:
    """
    Represents the inputs of the SLU API.
    """

    utterances: List[Utterance] = attr.ib(kw_only=True)
    """
    ASRs produce utterances. 
    Each utterance contains N-hypothesis. 
    Each hypothesis is a :code:`dict` with keys :code:`"transcript"` and :code:`"confidence"`.
    """

    reference_time: Optional[int] = attr.ib(default=None, kw_only=True)
    """
    The time that should be used for parsing relative time values.
    This is a Unix timestamp in seconds.
    utils/datetime.py has :ref:`make_unix_ts <make_unix_ts>` to help convert 
    a date in ISO 8601 format to unix ms timestamp.
    """

    latent_entities: bool = attr.ib(default=False, kw_only=True, converter=bool)
    """
    A switch to turn on/off production of latent entities via Duckling API
    If you need to parse "4" as 4 am or 4 pm. Note the absence of "am" or "pm" in the utterance.
    Then you might need this to be set as :code:`True`.
    It may be helpful to keep it :code:`False` unless clearly required. 
    """

    transcripts: List[str] = attr.ib(default=None)
    """
    A derived attribute. We cross product each utterance and return a list of strings.
    We use this to :ref:`normalize <normalize>` utterances.
    """

    clf_feature: Optional[List[str]] = attr.ib(  # type: ignore
        kw_only=True,
        factory=list,
        validator=attr.validators.optional(attr.validators.instance_of(list)),
    )
    """
    Placeholder for the features of an intent classifier.
    
    .. code-block:: python

        ["<s> I want to book a flight </s> <s> I want to book flights </s> <s> I want to book a flight to Paris </s>"]
    """

    lang: str = attr.ib(
        default="en", kw_only=True, validator=attr.validators.instance_of(str)
    )
    """
    Expected language of the input. This is needed for language dependent plugins.
    These are present in https://en.wikipedia.org/wiki/List_of_ISO_639-1_codes
    for "English" the code is "en". 
    """

    locale: str = attr.ib(
        default="en_IN",
        kw_only=True,
        validator=attr.validators.optional(attr.validators.instance_of(str)),  # type: ignore
    )
    """
    The locale identifier consists of at least a language code and a country/region code.
    We keep "en_IN" as our default. This is used by Duckling for parsing patterns as per the 
    locale. If locale is missing i.e. None, we may fallback to :code:`lang` instead.
    """

    timezone: str = attr.ib(
        default="UTC",
        kw_only=True,
        validator=attr.validators.optional(attr.validators.instance_of(str)),  # type: ignore
    )
    """
    Timezones from https://en.wikipedia.org/wiki/List_of_tz_database_time_zones
    Used by duckling or any other date/time parsing plugins.
    """

    slot_tracker: Optional[List[Dict[str, Any]]] = attr.ib(
        default=None,
        kw_only=True,
        validator=attr.validators.optional(attr.validators.instance_of(list)),
    )
    """
    This data structure tracks the slots that were filled in previous turns.
    This may come handy if we want to filter or reduce entities depending on our history.
    We use this in our :ref:`CombineDateTimeOverSlots <CombineDateTimeOverSlots>` plugin.

    .. code-block:: python

        [{
            "name": "_callback_",               # the intent name
            "slots": [{
                "name": "callback_datetime",    # the slot name
                "type": [                       # can fill entities of these types.
                    "time",
                    "date",
                    "datetime"
                ],
                "values": [{                    # entities that were filled previously
                    "alternative_index": 0,
                    "body": "tomorrow",
                    "entity_type": "date",
                    "grain": "day",
                    "parsers": ["duckling"],
                    "range": {
                        "end": 8,
                        "start": 0
                    },
                    "score": None,
                    "type": "value",
                    "value": "2021-10-15T00:00:00+05:30"
                }]
            }]
        }]
    """

    current_state: Optional[str] = attr.ib(
        default=None,
        kw_only=True,
        validator=attr.validators.optional(attr.validators.instance_of(str)),
    )
    """
    Points at the active state (or node) within the conversation graph.
    """

    previous_intent: Optional[str] = attr.ib(
        default=None,
        kw_only=True,
        validator=attr.validators.optional(attr.validators.instance_of(str)),
    )
    """
    The name of the intent that was predicted by the classifier in (exactly) the previous turn.
    """
    history: Optional[List[Dict[str, Any]]] = attr.ib(default=None, kw_only=True)

    def __attrs_post_init__(self) -> None:
        try:
            object.__setattr__(self, "transcripts", normalize(self.utterances))
        except TypeError:
            ...

    @reference_time.validator  # type: ignore
    def _check_reference_time(
        self, attribute: attr.Attribute, reference_time: Optional[int]  # type: ignore
    ) -> None:
        if reference_time is None:
            return
        if not isinstance(reference_time, int):
            raise TypeError(f"{attribute.name} must be an integer.")
        if not is_unix_ts(reference_time):
            raise ValueError(
                f"{attribute.name} must be a unix timestamp but got {reference_time}."
            )

    def json(self) -> Dict[str, Any]:
        """
        Serialize `Input`_ to a JSON-like dict.

        :return: A dictionary that represents an `Input`_ instance.
        :rtype: Dict[str, Any]
        """
        return attr.asdict(self)

    @classmethod
    def from_dict(cls, d: Dict[str, Any], reference: Optional[Input] = None) -> Input:
        """
        Create a new `Input`_ instance from a dictionary.

        :param d: A dictionary such that keys are a subset of `Input`_ attributes.
        :type d: Dict[str, Any]
        :param reference: An existing `Input`_ instance., defaults to None
        :type reference: Optional[Input], optional
        :return: A new `Input`_ instance.
        :rtype: Input
        """
        if reference:
            return attr.evolve(reference, **d)
        return attr.evolve(cls(utterances=d["utterances"]), **d)  # type: ignore



from datetime import datetime
from typing import Callable, Union

import pytz


def is_unix_ts(ts: int) -> bool:
    """
    Check if the input is a unix timestamp.

    :param ts: A unix timestamp (13-digit).
    :type ts: int
    :return: True if :code:`ts` is a unix timestamp, else False.
    :rtype: bool
    """
    try:
        datetime.fromtimestamp(ts / 1000)
        return True
    except ValueError:
        return False


def dt2timestamp(date_time: datetime) -> int:
    """
    Converts a python datetime object to unix-timestamp.

    :param date_time: An instance of datetime.
    :type date_time: datetime
    :return: Unix timestamp integer.
    :rtype: int
    """
    return int(date_time.timestamp() * 1000)


def make_unix_ts(tz: str = "UTC") -> Callable[[str], int]:
    """
    Convert date in ISO 8601 format to unix ms timestamp.

    .. _make_unix_ts:

    .. ipython ::

        In [1]: from dialogy.utils.datetime import make_unix_ts

        In [2]: ts = make_unix_ts("Asia/Kolkata")("2022-02-07T19:39:39.537827")

        In [3]: ts == 1644241599537

    :param tz: A timezone string, defaults to "UTC"
    :type tz: Optional[str], optional
    :return: A callable that converts a date in ISO 8601 format to unix ms timestamp.
    :rtype: Callable[[str], int]
    """

    def make_tz_aware(date_string: Union[str, int]) -> int:
        """
        Convert date in ISO 8601 format to unix ms timestamp.

        :param date_string: A date in ISO 8601 format.
        :type date_string: str
        :return: A unix timestamp (13 digit).
        :rtype: int
        """
        if isinstance(date_string, int):
            return date_string
        dt = datetime.fromisoformat(date_string)
        if dt.tzinfo is None and tz not in pytz.all_timezones:
            raise ValueError(f"Invalid timezone {tz} pick from:\n{pytz.all_timezones}")

        if dt.tzinfo is None:
            dt = dt.replace(tzinfo=pytz.timezone(tz))
        return dt2timestamp(dt)

    return make_tz_aware


def unix_ts_to_datetime(reference_time: int, timezone: str = "UTC") -> datetime:
    return datetime.fromtimestamp(reference_time / 1000, pytz.timezone(timezone))



"""
.. _ListEntityPlugin:

Regex Search
-------------

We have often seen certain keywords that gain significance in an SLU project. These keywords are 
easy to extract via patterns and are often used to create entities. The :ref:`ListEntityPlugin<ListEntityPlugin>`
helps in this task, it requires a pattern-map, we call it :code:`candidates`.


.. ipython::

    In [1]: from dialogy.workflow import Workflow
       ...: from dialogy.plugins import ListEntityPlugin
       ...: from dialogy.base import Input

    In [2]: candidates = {
       ...:      "colours": {
       ...:          "red": ["red", "crimson", "ruby", "raspberry"],
       ...:          "blue": ["blue", "azure", "indigo", "navy"],
       ...:          "green": ["green", "emerald", "jade", "teal"],
       ...:      }, 
       ...:    }

    In [2]: list_entity_plugin = ListEntityPlugin(
       ...:   candidates=candidates,
       ...:   style="regex",
       ...:   dest="output.entities"
       ...: )
       ...: workflow = Workflow([list_entity_plugin])
       ...: _, output = workflow.run(Input(utterances="I want an emerald green shirt"))

    In [3]: output

Spacy NER
-----------

We also allow using `spacy's NER <https://spacy.io/usage/spacy-101#annotations-ner>`_. This has to be passed within the
:code:`spacy_nlp` attribute.

.. ipython::

    In [1]: import spacy
       ...: from dialogy.workflow import Workflow
       ...: from dialogy.plugins import ListEntityPlugin
       ...: from dialogy.base import Input

    In [2]: nlp = spacy.load("en_core_web_sm")

    In [3]: list_entity_plugin = ListEntityPlugin(
       ...:   spacy_nlp=nlp,
       ...:   style="spacy",
       ...:   dest="output.entities"
       ...: )
       ...: workflow = Workflow([list_entity_plugin])
       ...: _, output = workflow.run(Input(utterances="Need a place to stay in New Delhi."))

    In [3]: output
"""
import re
from pprint import pformat
from typing import Any, Dict, List, Optional, Tuple

import pandas as pd
from loguru import logger
from tqdm import tqdm

from dialogy import constants as const
from dialogy.base import Guard, Input, Output, Plugin
from dialogy.base.entity_extractor import EntityScoringMixin
from dialogy.types import BaseEntity, KeywordEntity

Text = str
Label = str
Span = Tuple[int, int]
Value = str
MatchType = List[Tuple[Text, Label, Value, Span]]


class ListEntityPlugin(EntityScoringMixin, Plugin):
    """
     A :ref:`Plugin<AbstractPlugin>` for extracting entities using spacy or a list of regex patterns.

     .. attention:

        This class will undergo a series of refactoring changes. FWIW, :ref:`ListSearchPlugin<ListSearchPlugin>`
        is more more performant in terms of entity capture rates but not as responsive. :code:`ListEntityPlugin`
        is fast. So make choices with bearing this in mind.

    :param style: One of ["regex", "spacy"]
    :type style: Optional[str]
    :param candidates: Required if style is "regex", this is a :code:`dict` that shows a mapping of entity
        values and their patterns.
    :type candidates: Optional[Dict[str, List[str]]]
    :param spacy_nlp: Required if style is "spacy", requires is a
        `spacy model <https://spacy.io/usage/spacy-101#annotations-ner>`_.
    :type spacy_nlp: Any
    :param labels: Required if style is "spacy". If there is a need to extract only a few labels from all the other
        `available labels <https://github.com/explosion/spaCy/issues/441#issuecomment-311804705>`_.
    :type labels: Optional[List[str]]
    :param debug: A flag to set debugging on the plugin methods
    :type debug: bool
    """

    def __init__(
        self,
        style: Optional[str] = None,
        candidates: Optional[Dict[str, Dict[str, List[Any]]]] = None,
        spacy_nlp: Any = None,
        dest: Optional[str] = None,
        guards: Optional[List[Guard]] = None,
        labels: Optional[List[str]] = None,
        threshold: Optional[float] = None,
        input_column: str = const.ALTERNATIVES,
        output_column: Optional[str] = None,
        use_transform: bool = True,
        flags: re.RegexFlag = re.I | re.U,
        debug: bool = False,
    ):
        super().__init__(
            dest=dest,
            guards=guards,
            debug=debug,
            input_column=input_column,
            output_column=output_column,
            use_transform=use_transform,
        )
        self.__style_search_map = {
            const.SPACY: self.ner_search,
            const.REGEX: self.regex_search,
        }

        self.style = style or const.REGEX
        self.labels = labels
        self.threshold = threshold
        self.keywords = None
        self.spacy_nlp = spacy_nlp
        self.compiled_patterns: Dict[str, Dict[str, List[Any]]] = {}
        self.flags = flags

        if self.style == const.REGEX:
            self._parse(candidates)

    def _parse(self, candidates: Optional[Dict[str, Dict[str, List[Any]]]]) -> None:
        """
        Pre compile regex patterns to speed up runtime evaluation.

        This method's search will still be slow depending on the list of patterns.

        :param candidates: A map for entity types and their pattern list.
        :type candidates: Optional[Dict[str, List[str]]]
        :return: None
        :rtype: NoneType
        """
        logger.debug(
            pformat(
                {
                    "style": self.style,
                    "candidates": candidates,
                }
            )
        )
        if not isinstance(candidates, dict):
            raise TypeError(
                'Expected "candidates" to be a Dict[str, List[str]]'
                f" but {type(candidates)} was found."
            )

        if not candidates:
            raise ValueError(
                'Expected "candidates" to be a Dict[str, List[str]]'
                f" but {candidates} was found."
            )

        if self.style not in self.__style_search_map:
            raise ValueError(
                f"Expected style to be one of {list(self.__style_search_map.keys())}"
                f' but "{self.style}" was found.'
            )
        self.compiled_patterns = {}

        if self.style == const.REGEX:
            for entity_type, entity_value_dict in candidates.items():
                for entity_value, entity_patterns in entity_value_dict.items():
                    patterns = [
                        re.compile(pattern, flags=self.flags)
                        for pattern in entity_patterns
                    ]
                    if not self.compiled_patterns:
                        self.compiled_patterns = {entity_type: {entity_value: patterns}}
                    elif entity_type in self.compiled_patterns:
                        self.compiled_patterns[entity_type][entity_value] = patterns
                    else:
                        self.compiled_patterns[entity_type] = {entity_value: patterns}

        logger.debug("compiled patterns")
        logger.debug(self.compiled_patterns)

    def _search(self, transcripts: List[str]) -> List[MatchType]:
        """
        Search for tokens in a list of strings.

        :param transcripts: A list of transcripts, :code:`List[str]`.
        :type transcripts: List[str]
        :return: Token matches with the transcript.
        :rtype: List[MatchType]
        """
        logger.debug(f"style: {self.style}")
        logger.debug("transcripts")
        logger.debug(transcripts)
        search_fn = self.__style_search_map.get(self.style)
        if not search_fn:
            raise ValueError(
                f"Expected style to be one of {list(self.__style_search_map.keys())}"
                f' but "{self.style}" was found.'
            )
        token_list = [search_fn(transcript) for transcript in transcripts]
        return token_list

    def get_entities(self, transcripts: List[str]) -> List[BaseEntity]:
        """
        Parse entities using regex and spacy ner.

        :param transcripts: A list of strings within which to search for entities.
        :type transcripts: List[str]
        :return: List of entities from regex matches or spacy ner.
        :rtype: List[KeywordEntity]
        """
        matches_on_transcripts = self._search(transcripts)
        logger.debug(matches_on_transcripts)
        entities: List[BaseEntity] = []

        for i, matches_on_transcript in enumerate(matches_on_transcripts):
            for text, label, value, span in matches_on_transcript:
                entity_dict = {
                    const.RANGE: {
                        const.START: span[0],
                        const.END: span[1],
                    },
                    const.BODY: text,
                    const.DIM: label,
                    const.SCORE: 0,
                    const.ALTERNATIVE_INDEX: i,
                    const.LATENT: False,
                    "__group": f"{label}_{text}",
                    const.TYPE: label,
                    const.ENTITY_TYPE: label,
                    const.VALUE: value,
                    const.VALUES: [{const.VALUE: value}],
                }

                del entity_dict["__group"]
                entity_ = KeywordEntity.from_dict(entity_dict)
                entity_.add_parser(self)
                entities.append(entity_)

        logger.debug("Parsed entities")
        logger.debug(entities)

        aggregated_entities = self.entity_consensus(entities, len(transcripts))
        return self.apply_filters(aggregated_entities)

    def utility(self, input: Input, _: Output) -> Any:
        transcripts = input.transcripts
        return self.get_entities(transcripts)  # pylint: disable=no-value-for-parameter

    def ner_search(self, transcript: str) -> MatchType:
        """
        Wrapper over spacy's ner search.

        :param transcript: A string to search entities within.
        :type transcript: str
        :return: NER parsing via spacy.
        :rtype: MatchType
        """
        if not self.spacy_nlp:
            raise ValueError(
                "Expected spacy_nlp to be a spacy"
                f" instance but {self.spacy_nlp} was found."
            )
        parsed_transcript = self.spacy_nlp(transcript)

        if self.labels:
            return [
                (
                    token.text,
                    token.label_,
                    token.text,
                    (
                        transcript.index(token.text),
                        transcript.index(token.text) + len(token.text),
                    ),
                )
                for token in parsed_transcript.ents
                if token.label_ in self.labels
            ]
        else:
            return [
                (
                    token.text,
                    token.label_,
                    token.text,
                    (
                        transcript.index(token.text),
                        transcript.index(token.text) + len(token.text),
                    ),
                )
                for token in parsed_transcript.ents
            ]

    def regex_search(self, transcript: str) -> MatchType:
        """
        Wrapper over regex searches.

        :param transcript: A string to search entities within.
        :type transcript: str
        :return: regex parsing via spacy.
        :rtype: MatchType
        """
        if not self.compiled_patterns:
            raise TypeError(
                "Expected compiled_patterns to be "
                f"Dict[str, List[str]] but {type(self.compiled_patterns)} was found."
            )

        entity_tokens = []
        for entity_type, entity_value_dict in self.compiled_patterns.items():
            for entity_value, entity_patterns in entity_value_dict.items():
                for pattern in entity_patterns:
                    matches = pattern.search(transcript)
                    if matches:
                        entity_tokens.append(
                            (matches.group(), entity_type, entity_value, matches.span())
                        )
        logger.debug(entity_tokens)
        return entity_tokens

    def transform(self, training_data: pd.DataFrame) -> pd.DataFrame:
        """
        Transform training data.

        :param training_data: Training data.
        :type training_data: pd.DataFrame
        :return: Transformed training data.
        :rtype: pd.DataFrame
        """
        if not self.use_transform:
            return training_data

        logger.debug(f"Transforming dataset via {self.__class__.__name__}")
        training_data = training_data.copy()
        if self.output_column not in training_data.columns:
            training_data[self.output_column] = None

        logger.disable("dialogy")
        for i, row in tqdm(training_data.iterrows(), total=len(training_data)):
            transcripts = self.make_transform_values(row[self.input_column])
            entities = self.get_entities(transcripts)
            is_empty_series = isinstance(row[self.output_column], pd.Series) and (
                row[self.output_column].isnull()
            )
            is_row_nonetype = row[self.output_column] is None
            is_row_nan = pd.isna(row[self.output_column])
            if is_empty_series or is_row_nonetype or is_row_nan:
                training_data.at[i, self.output_column] = entities
            else:
                training_data.at[i, self.output_column] = (
                    row[self.output_column] + entities
                )
        logger.enable("dialogy")
        return training_data



"""
.. _list_Search_plugin_

Module needs refactor. We are currently keeping all strategies bundled as methods as opposed to SearchStrategyClasses.

Within dialogy, we extract entities using Duckling, Pattern lists and Spacy. We can ship individual plugins but at the
same time, the difference is just configuration of each of these tools/services. There is another difference of
intermediate structure that the DucklingPlugin expects. We need to prevent the impact of the structure from affecting
all other entities. So that their :code:`from_dict(...)` methods are pristine and involve no shape hacking.
"""
import re
from typing import Any, Dict, List, Optional, Tuple

import stanza
from loguru import logger
from thefuzz import fuzz

from dialogy import constants as const
from dialogy.base import Guard, Input, Output, Plugin
from dialogy.base.entity_extractor import EntityScoringMixin
from dialogy.types import BaseEntity, KeywordEntity

Text = str
Label = str
Span = Tuple[int, int]
Value = str
Score = float
MatchType = List[Tuple[Text, Label, Value, Span, Score]]  # adding score for each entity


class ListSearchPlugin(EntityScoringMixin, Plugin):
    """
     A :ref:`Plugin<plugin>` for extracting entities using spacy or a list of regex patterns.

     .. attention:

        This class will undergo a series of refactoring changes. FWIW, :ref:`ListSearchPlugin<ListSearchPlugin>`
        is more more performant in terms of entity capture rates but not as responsive. :code:`ListEntityPlugin`
        is fast. So make choices with bearing this in mind.

    .. _ListSearchPlugin:

    :param style: One of ["regex", "spacy"]
    :type style: Optional[str]
    :param candidates: Required if style is "regex", this is a :code:`dict` that shows a mapping of entity
        values and their patterns.
    :type candidates: Optional[Dict[str, List[str]]]
    :param spacy_nlp: Required if style is "spacy", this is a
        `spacy model <https://spacy.io/usage/spacy-101#annotations-ner>`_.
    :type spacy_nlp: Any
    :param labels: Required if style is "spacy". If there is a need to extract only a few labels from all the other
        `available labels <https://github.com/explosion/spaCy/issues/441#issuecomment-311804705>`_.
    :type labels: Optional[List[str]]
    :param debug: A flag to set debugging on the plugin methods
    :type debug: bool
    """

    def __init__(
        self,
        fuzzy_dp_config: Dict[Any, Any],  # parsed yaml file
        threshold: Optional[float] = None,
        dest: Optional[str] = None,
        guards: Optional[List[Guard]] = None,
        input_column: str = const.ALTERNATIVES,
        output_column: Optional[str] = None,
        use_transform: bool = True,
        flags: re.RegexFlag = re.I | re.U,
        debug: bool = False,
        fuzzy_threshold: Optional[float] = 0.1,
    ):
        super().__init__(
            dest=dest,
            guards=guards,
            debug=debug,
            input_column=input_column,
            output_column=output_column,
            use_transform=use_transform,
        )
        self.__style_search_map = {
            const.FUZZY_DP: self.get_fuzzy_dp_search,
        }

        self.style = (
            const.FUZZY_DP
        )  # which search algo will be used: [regex, spacy, fuzzy]
        self.keywords = None
        self.flags = flags
        self.threshold = threshold
        """
        Parameters for Fuzzy Dependency Parser defined below
        """

        # ensuring stanza models are downloaded
        stanza.download("en")
        stanza.download("hi")

        self.fuzzy_dp_config = fuzzy_dp_config
        self.entity_dict: Dict[Any, Any] = {}
        self.entity_types: Dict[Any, Any] = {}
        self.nlp: Dict[Any, Any] = {}
        self.fuzzy_threshold = fuzzy_threshold

        if self.style == const.FUZZY_DP:
            self.fuzzy_init()

    def fuzzy_init(self) -> None:
        """
        Initializing the parameters for fuzzy dp search with their values

        """
        valid_langs = ["hi", "en"]
        for lang_code in self.fuzzy_dp_config.keys():
            if lang_code not in valid_langs:
                raise ValueError(
                    f"Provided language {lang_code} is not supported by this method at present"
                )
            self.entity_dict[lang_code] = self.fuzzy_dp_config[lang_code]
            self.entity_types[lang_code] = list(self.entity_dict[lang_code].keys())
            self.nlp[lang_code] = stanza.Pipeline(
                lang=lang_code, tokenize_pretokenized=True
            )

    def _search(self, transcripts: List[str], lang: str) -> List[MatchType]:
        """
        Search for tokens in a list of strings.

        :param transcripts: A list of transcripts, :code:`List[str]`.
        :type transcripts: List[str]
        :return: Token matches with the transcript.
        :rtype: List[MatchType]
        """
        logger.debug(f"style: {self.style}")
        logger.debug("transcripts")
        logger.debug(transcripts)
        search_fn = self.get_fuzzy_dp_search

        token_list = [search_fn(transcript, lang=lang) for transcript in transcripts]
        return token_list

    def search_regex(
        self,
        query: str,
        entity_type: str = "",
        entity_patterns: List[str] = [""],
        match_dict: Dict[Any, Any] = {},
    ) -> Tuple[Text, Label, Value, Span, Score]:
        max_length = 0
        final_match = None

        for pattern in entity_patterns:
            result = re.search(pattern, query)
            if result:
                match_value = match_dict[result.group()]
                match_len = len(match_value)
                if match_len > max_length:
                    match_text = result.group()
                    max_length = match_len
                    final_match = match_value
                    match_span = result.span()
        if final_match:
            return (match_text, entity_type, final_match, match_span, 1.0)
        return ("", entity_type, "", (0, 0), 0.0)

    def dp_search(
        self,
        query: str,
        nlp: Any,
        entity_type: str = "",
        entity_patterns: List[str] = [""],
        match_dict: Dict[Any, Any] = {},
    ) -> Tuple[Text, Label, Value, Span, Score]:

        sentence = nlp(query).sentences[0]
        value = ""
        pos_tags = ["PROPN", "NOUN", "ADP"]
        result_dict = {}
        for word in sentence.words:
            if word.upos in pos_tags:
                if value == "":
                    span_start = word.start_char
                span_end = word.end_char

                """
                joining individual tokens that together are the real entity,
                Since we are dealing with Multi-Word entities here

                """
                value = value + str(word.text) + " "
        if value != "":
            for pattern in entity_patterns:
                val = fuzz.ratio(pattern, value) / 100
                if val > self.fuzzy_threshold:
                    match_value = match_dict[pattern]
                    result_dict[match_value] = val
            if result_dict:
                match_output = max(result_dict, key=lambda x: result_dict[x])
                match_score = result_dict[match_output]

                return (
                    value,
                    entity_type,
                    match_output,
                    (span_start, span_end),
                    match_score,
                )
        return (value, entity_type, "", (0, 0), 0.0)

    # new method based on experiments done during development of channel parser
    def get_fuzzy_dp_search(self, transcript: str, lang: str = "") -> MatchType:
        """
        Search for Entity in transcript from a defined List Search space
        :param transcripts : A list of transcripts, :code:`List[str]`.
        :param lang : Language code of the transcript :code str
        :return: Token matches with the transcript.
        :rtype: List[MatchType]

        """
        match = []
        query = transcript

        entity_patterns = {}
        entity_match_dict = {}
        for entity_type in self.entity_types[lang]:
            entity_patterns[entity_type] = list(
                self.entity_dict[lang][entity_type].keys()
            )
            entity_match_dict[entity_type] = self.entity_dict[lang][entity_type]
            match_entity = self.search_regex(
                query,
                entity_type,
                entity_patterns[entity_type],
                entity_match_dict[entity_type],
            )

            if match_entity[0] == "":
                match_entity = self.dp_search(
                    query,
                    self.nlp[lang],
                    entity_type,
                    entity_patterns[entity_type],
                    entity_match_dict[entity_type],
                )
            match.append(match_entity)

        return match

    def get_entities(self, transcripts: List[str], lang: str) -> List[BaseEntity]:
        """
        Parse entities using regex and spacy ner.

        :param transcripts: A list of strings within which to search for entities.
        :type transcripts: List[str]
        :return: List of entities from regex matches or spacy ner.
        :rtype: List[KeywordEntity]
        """
        matches_on_transcripts = self._search(transcripts, lang)
        logger.debug(matches_on_transcripts)
        entities: List[BaseEntity] = []

        for i, matches_on_transcript in enumerate(matches_on_transcripts):
            for text, label, value, span, score in matches_on_transcript:
                if score == 0.0:
                    continue
                entity_dict = {
                    const.RANGE: {
                        const.START: span[0],
                        const.END: span[1],
                    },
                    const.BODY: text,
                    const.DIM: label,
                    const.SCORE: 0,
                    const.ALTERNATIVE_INDEX: i,
                    const.LATENT: False,
                    "__group": f"{label}_{text}",
                    const.TYPE: label,
                    const.ENTITY_TYPE: label,
                    const.VALUE: value,
                    const.VALUES: [{const.VALUE: value}],
                }

                del entity_dict["__group"]
                entity_ = KeywordEntity.from_dict(entity_dict)
                entity_.add_parser(self)
                entities.append(entity_)

        logger.debug("Parsed entities")
        logger.debug(entities)
        aggregated_entities = self.entity_consensus(entities, len(transcripts))
        return self.apply_filters(aggregated_entities)

    def utility(self, input_: Input, _: Output) -> Any:
        return self.get_entities(
            input_.transcripts, input_.lang
        )  # pylint: disable=no-value-for-parameter



"""
Module provides types

Import Types:
    - Alternative
    - Utterance
"""
from typing import Dict, List, Optional, Union

Transcript = str
Alternative = Dict[str, Union[str, Optional[float]]]
Utterance = List[Alternative]



"""
The :ref:`Workflow<WorkflowClass>` represents a black box for an SLU microservice, and its terminals, :ref:`input <Input>` and :ref:`output <Output>`
represent request response structure of the SLU API.

.. note::

    The :ref:`Input<Input>` has some extra properties that aren't part of the SLU API.
    We have kept reserves for intermediates, and flattened as most other nested inputs 
    aren't required and add to bloat.

.. attention::

    :ref:`Input<Input>` and :ref:`Output<Output>` instances are immutable.

What does it do?
----------------

Produce the response for a single turn of a conversation. It recieves (not an exhaustive list):

- Utterance.
- Timezone
- Language code
- Locale
- Current State of the conversation
- Slot tracker


as :ref:`inputs<Input>`, parses these, produces a few intermediates and finally returns the :ref:`intent <Intent>` 
and optionally :ref:`entities <BaseEntity>`. Details will be documented within the specific plugins.

.. mermaid::

    classDiagram
        direction TB
        Workflow --> "1" Input: has
        Workflow --> "1" Output: has
        Workflow --> "many" Plugin: has

        class Workflow {
            Workflow: +Input input
            Workflow: +Output output
            Workflow: +set(path: str, value: Any)
            Workflow: +run(input: Input)
        }

        class Input {
            +List~Utterance~ utterance
            +int reference_time
            +str current_state
            +str lang
            +str locale
            +dict slot_tracker
            +str timezone
            +dict json()
        }

        class Output {
            +List~Intent~ intents
            +List~BaseEntity~ entities
            +dict json()
        }

        <<abstract>> Plugin
        class Plugin {
            +str dest
            *any utility(input: Input, output: Output)
        }



How does it do it?
------------------

- A workflow contains a sequence of :ref:`plugins <AbstractPlugin>`. The sequence is important.
- The sequence describes the order in which :ref:`plugins <AbstractPlugin>` are run.
- The plugins can save their output within the workflow's :ref:`input<Input>` or :ref:`output<Output>`.
- After execution of all plugins, the :ref:`workflow <WorkflowClass>` returns a pair of serialized :ref:`input<Input>` and :ref:`output<Output>`.


.. mermaid::

    flowchart TB

        subgraph Workflow

            subgraph input
                utterance
                reference_time
            end

            subgraph output
                intents
                entities
            end

            input --> plugin
            output --> plugin
            plugin -->|update| input
            plugin -->|update| output
            plugin -->|foreach| plugin

        end
        output --> output_json
        input --> input_json

        output_json --> workflow_output
        input_json --> workflow_output

        subgraph Response
            output_json
        end

"""
from __future__ import annotations

import copy
import time
from threading import Lock
from typing import Any, Dict, List, Optional, Tuple

import attr
import pandas as pd

from dialogy import constants as const
from dialogy.base.input import Input
from dialogy.base.output import Output
from dialogy.base.plugin import Plugin
from dialogy.utils.logger import logger


@attr.s
class Workflow:
    """
    SLU API blackbox.

    .. _WorkflowClass:
    """

    plugins = attr.ib(
        factory=list,
        type=List[Plugin],
        validator=attr.validators.instance_of(list),
    )
    """
    List of :ref:`plugins <AbstractPlugin>`.
    """

    input: Optional[Input] = attr.ib(
        default=None,
        kw_only=True,
        validator=attr.validators.optional(attr.validators.instance_of(Input)),
    )
    """
    Represents the SLU API request object. A few nested properties and intermediates are flattened out
    for readability.
    """

    output: Optional[Output] = attr.ib(
        default=None,
        kw_only=True,
        validator=attr.validators.optional(attr.validators.instance_of(Output)),
    )
    """
    Represents the SLU API response object.
    """

    debug = attr.ib(
        type=bool, default=False, validator=attr.validators.instance_of(bool)
    )
    """
    Switch on/off the debug logs for the workflow.
    """
    lock: Lock
    """
    A :ref:`workflow <WorkflowClass>` isn't thread-safe by itself. We need locks
    to prevent race conditions.
    """
    NON_SERIALIZABLE_FIELDS = [const.PLUGINS, const.DEBUG, const.LOCK]

    def __attrs_post_init__(self) -> None:
        """
        Post init hook.
        """
        self.__reset()
        self.lock = Lock()

    def __reset(self) -> None:
        """
        Use this method to keep workflow-io in the same format as expected.
        """
        self.input = None
        self.output = Output()

    def set(self, path: str, value: Any, replace: bool = False) -> Workflow:
        """
        Set attribute path with value.

        This method (re)-sets the input or output object without losing information
        from previous instances.

        :param path: A '.' separated attribute path.
        :type path: str
        :param value: A value to set.
        :type value: Any
        :return: This instance
        :rtype: Workflow
        """
        dest, attribute = path.split(".")

        if dest == "input":
            self.input = Input.from_dict({attribute: value}, reference=self.input)
        elif dest == "output" and isinstance(value, list):

            if replace:
                self.output = Output.from_dict(
                    {attribute: value}, reference=self.output
                )
            else:
                if attribute == const.INTENTS:
                    previous_value = self.output.intents  # type: ignore
                elif attribute == const.ENTITIES:
                    previous_value = self.output.entities  # type: ignore
                else:
                    raise ValueError(
                        f"Unknown attribute {attribute} tracked on Output."
                    )

                combined_value = sorted(
                    previous_value + value,
                    key=lambda parse: parse.score or 0,
                    reverse=True,
                )
                self.output = Output.from_dict(
                    {attribute: combined_value}, reference=self.output
                )

        elif dest == "output":
            raise ValueError(f"{value=} should be a List[Intent] or List[BaseEntity].")
        else:
            raise ValueError(f"{path} is not a valid path.")
        return self

    def execute(self) -> Workflow:
        """
        Update input, output attributes.

        We iterate through pre/post processing functions and update the input and
        output attributes of the class. It is expected that pre-processing functions
        would modify the input, and post-processing functions would modify the output.
        """
        history = {}
        for plugin in self.plugins:
            if not callable(plugin):
                raise TypeError(f"{plugin=} is not a callable")

            # logs are available only when debug=False during class initialization
            if self.debug:
                history = {
                    "plugin": plugin,
                    "before": {
                        "input": self.input,
                        "output": self.output,
                    },
                }
            start = time.perf_counter()
            plugin(self)
            end = time.perf_counter()
            # logs are available only when debug=False during class initialization
            if self.debug:
                history["after"] = {
                    "input": self.input,
                    "output": self.output,
                }
                history["perf"] = round(end - start, 4)
            if history:
                logger.debug(history)
        return self

    def run(self, input_: Input) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """
        .. _workflow_run:

        Get final results from the workflow.

        The current workflow exhibits the following simple procedure:
        pre-processing -> inference -> post-processing.
        """
        with self.lock:
            self.input = input_
            return self.execute().flush()

    def flush(self) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """
        Reset :code:`workflow.input` and :code:`workflow.output`.
        """
        if self.input is None or self.output is None:
            return {}, {}
        input_ = copy.deepcopy(self.input.json())
        output = copy.deepcopy(self.output.json())
        self.__reset()
        return input_, output

    def json(self) -> Dict[str, Any]:
        """
        Represent the workflow as a python dict.

        :rtype: Dict[str, Any]
        """
        return attr.asdict(
            self,
            filter=lambda attribute, _: attribute.name
            not in self.NON_SERIALIZABLE_FIELDS,
        )

    def train(self, training_data: pd.DataFrame) -> Workflow:
        """
        Train all the plugins in the workflow.

        Plugin's have a no-op train method by default. The one's that do require training
        should override this method. All trainable plugins manage their data validation
        (in case the data isn't suitable for them) and saving/loading artifacts.

        :param training_data: [description]
        :type training_data: pd.DataFrame
        """
        for plugin in self.plugins:
            plugin.train(training_data)
            transformed_data = plugin.transform(training_data)
            if transformed_data is not None:
                training_data = transformed_data
        return self


47



import argparse
import os
import shutil

from copier import copy

import dialogy.constants as const


def manage_project(
    destination_path: str,
    template: str = const.DEFAULT_PROJECT_TEMPLATE,
    namespace: str = const.DEFAULT_NAMESPACE,
    use_master: bool = False,
    pretend: bool = False,
    is_update: bool = False,
) -> None:
    """
    Create a new project using scaffolding from an existing template.

    This function uses `copier's <https://copier.readthedocs.io/en/stable/>`_ `copy <https://copier.readthedocs.io/en/stable/#quick-usage>`_ to use an existing template.

    An example template is `here: <https://github.com/Vernacular-ai/dialogy-template-simple-transformers>`_.

    :param destination_path: The directory where the scaffolding must be generated, creates a dir if missing but aborts if there are files already in the specified location.
    :type destination_path: str
    :param template: Scaffolding will be generated using a copier template project. This is the link to the project.
    :type template: str
    :param namespace: The user or the organization that supports the template, defaults to "vernacular-ai"
    :type namespace: str, optional
    :param vcs_ref: support for building from local git templates optionally, `--vcs` takes `"TAG"` or `"HEAD"`. defaults to `None`.
    :type vcs_ref: str, optional
    :return: None
    :rtype: NoneType
    """
    # to handle copier vcs associated git template building.
    if use_master:
        copy(
            template,
            destination_path,
            vcs_ref="HEAD",
            pretend=pretend,
            only_diff=is_update,
            force=is_update,
        )
    else:
        copy(
            f"gh:{namespace}/{template}.git",
            destination_path,
            only_diff=is_update,
            pretend=pretend,
            force=is_update,
        )

    return None


def project_cli(args: argparse.Namespace) -> None:
    """CLI for project command."""
    destination_path = project_name = args.project
    template_name = args.template
    namespace = args.namespace
    use_master = args.master
    is_update = args.command == "update"
    pretend = args.dry_run

    if (
        os.path.exists(destination_path)
        and os.listdir(destination_path)
        and not is_update
    ):
        raise RuntimeError("There are files on the destination path. Aborting !")

    if not os.path.exists(project_name):
        os.mkdir(destination_path)
        if pretend:
            shutil.rmtree(destination_path)

    manage_project(
        project_name,
        template=template_name,
        namespace=namespace,
        use_master=use_master,
        is_update=is_update,
        pretend=pretend,
    )



"""
This module exposes a command line utility to create project scaffolding. This creates a starting point for any project, an exhibit:

.. code-block:: bash
    :linenos:
    :emphasize-lines: 1

    dialogy create hello-world
    cd hello-world
    poetry install
    make lint

usage: dialogy [-h] [--template TEMPLATE] [--namespace NAMESPACE] [--vcs {HEAD,TAG}] {create,update} project

positional arguments:
  {create,update}
  project               A directory with this name will be created at the root of command invocation.

optional arguments:
  -h, --help            show this help message and exit
  --template TEMPLATE
  --namespace NAMESPACE
                        The github/gitlab user or organization name where the template project lies.
  --vcs {HEAD,TAG}      Download the template's latest tag (TAG) or master branch (HEAD).
"""
import argparse
from typing import Optional

import dialogy.constants as const
from dialogy.cli import project, workflow


def add_project_command_arguments(
    parser: argparse.ArgumentParser,
) -> argparse.ArgumentParser:
    parser.add_argument(
        "project",
        help="A directory with this name will be created at the root of command invocation.",
    )
    parser.add_argument("--template", default=const.DEFAULT_PROJECT_TEMPLATE)
    parser.add_argument(
        "--dry-run",
        help="Make no change to the directory structure.",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--namespace",
        help="The github/gitlab user or organization name where the template project lies.",
        default=const.DEFAULT_NAMESPACE,
    )
    parser.add_argument(
        "--master",
        help="Download the template's master branch (HEAD) instead of the latest tag.",
        action="store_true",
    )
    return parser


def add_workflow_command_arguments(
    parser: argparse.ArgumentParser, objective: str
) -> argparse.ArgumentParser:
    parser.add_argument(
        "module", help="The module that contains a Workflow (or subclass)."
    )
    parser.add_argument(
        "--fn",
        help="A function that returns a Workflow or its subclass's instance.",
        required=True,
    )
    parser.add_argument(
        "--data",
        help=f"The dataset (.csv) to be use for {objective.lower()}ing the workflow.",
        required=True,
    )
    parser.add_argument(
        "--project",
        help=f"The project to be used for {objective.lower()}ing.",
        default="slu",
    )
    parser.add_argument(
        "--lang",
        help=f"The language expected for {objective.lower()}ing.",
        default="all",
    )
    return parser


def project_command_parser(command_string: Optional[str]) -> argparse.Namespace:
    argument_parser = argparse.ArgumentParser()

    parser = argument_parser.add_subparsers(
        help="Dialogy project utilities.", dest="command"
    )
    create_parser = parser.add_parser(const.CREATE, help="Create a new project.")
    update_parser = parser.add_parser(
        const.UPDATE, help="Migrate an existing project to the latest template version."
    )
    create_parser = add_project_command_arguments(create_parser)
    update_parser = add_project_command_arguments(update_parser)
    train_workflow_parser = parser.add_parser(const.TRAIN, help="Train a workflow.")
    train_workflow_parser = add_workflow_command_arguments(
        train_workflow_parser, const.TRAIN
    )

    command = command_string.split() if command_string else None
    return argument_parser.parse_args(args=command)


def main(command_string: Optional[str] = None) -> None:
    """
    Create project scaffolding cli.
    """
    args = project_command_parser(command_string=command_string)
    if args.command in [const.CREATE, const.UPDATE]:
        project.project_cli(args)
    elif args.command in [const.TRAIN]:
        workflow.workflow_cli(args)



"""
Module provides utility functions for entities.

Import functions:
    - dict_traversal
    - validate_type
"""
from functools import reduce
from typing import Any, Dict, List, Tuple, Union


def traverse_dict(obj: Dict[Any, Any], properties: List[str]) -> Any:
    """
    Traverse a dictionary for a given list of properties.

    This is useful for traversing a deeply nested dictionary.
    Instead of recursion, we are using reduce to update the `dict`.
    Missing properties will lead to KeyErrors.

    .. ipython:: python
        :okexcept:

        from dialogy.utils import traverse_dict

        input_ = {
            "planets": {
                "mars": [{
                    "name": "",
                    "languages": [{
                        "beep": {"speakers": 11},
                    }, {
                        "bop": {"speakers": 30},
                    }]
                }]
            }
        }
        traverse_dict(input_, ["planets", "mars", 0 , "languages", 1, "bop"])

        # element with index 3 doesn't exist!
        traverse_dict(input_, ["planets", "mars", 0 , "languages", 3, "bop"])

    :param obj: The `dict` to traverse.
    :type obj: Dict[Any, Any]
    :param properties: List of properties to be parsed as a path to be navigated in the `dict`.
    :type properties: List[int]
    :return: A value within a deeply nested dict.
    :rtype: Any
    :raises KeyError: Missing property in the dictionary.
    :raises TypeError: Properties don't describe a path due to possible type error.
    """
    try:
        return reduce(lambda o, k: o[k], properties, obj)
    except KeyError as key_error:
        raise KeyError(
            f"Missing property {key_error} in {obj}. Check the types. Failed for path {properties}"
        ) from key_error
    except TypeError as type_error:
        raise TypeError(
            f"The properties aren't describing path within a dictionary. | {type_error}. Failed for path {properties}"
        ) from type_error


def validate_type(obj: Any, obj_type: Union[type, Tuple[type]]) -> None:
    """
    Raise TypeError on object type mismatch.

        This is syntatic sugar for instance type checks.

        The check is by exclusion of types. Wraps exception raising logic.

        :param obj: An object available for type assertion
        :type obj: Any
        :param obj_type: This must match the type of the object.
        :type obj_type: (Union[type, Tuple[type]])
        :return:
        :rtype:
        :raises TypeError: If the type `obj_type` doesn't match the type of `obj`.
    """
    if not isinstance(obj, obj_type):
        raise TypeError(f"{obj} should be a {obj_type}")



import argparse
import importlib
from typing import Any

import pandas as pd
from loguru import logger

import dialogy.constants as const
from dialogy.workflow import Workflow


def get_workflow(
    module: str, get_workflow_fn: str, purpose: str, **kwargs: Any
) -> Workflow:
    return getattr(importlib.import_module(module), get_workflow_fn)(purpose, **kwargs)


def train_workflow(args: argparse.Namespace) -> None:
    """
    Train a workflow.
    """
    module = args.module
    get_workflow_fn = args.fn
    data = args.data
    lang = args.lang
    project = args.project
    kwargs = {"lang": lang, "project": project}

    workflow: Workflow = get_workflow(module, get_workflow_fn, args.command, **kwargs)
    train_df = pd.read_csv(data)
    try:
        workflow.train(train_df)
    except AttributeError as error:
        logger.error(f"{workflow=} doesn't have a train method? 🤔")
        raise AttributeError from error


def workflow_cli(args: argparse.Namespace) -> None:
    """
    CLI entry point for workflows.

    args contain the following attributes:

    module      - We expect the Workflow or its subclass to be available at this path.
    fn          - A function that provides a Workflow or its subclass' instance.
    data        - Path to dataset (.csv).
    output_dir  - Path where artifacts are expected.
    """
    command = args.command
    try:
        if command == const.TRAIN:
            train_workflow(args)
    except ModuleNotFoundError as error:
        logger.error(
            f"Could not import module {args.module} is "
            "your virtual environment active? 🤔.\n"
        )
        raise ModuleNotFoundError from error
    except AttributeError as error:
        logger.error(f"Could not find {args.fn} in {args.module}? 🤔.\n{error}")
        raise AttributeError from error

    return None



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dialogy',
 'dialogy.base',
 'dialogy.base.entity_extractor',
 'dialogy.base.input',
 'dialogy.base.output',
 'dialogy.base.plugin',
 'dialogy.cli',
 'dialogy.constants',
 'dialogy.plugins',
 'dialogy.plugins.text',
 'dialogy.plugins.text.calibration',
 'dialogy.plugins.text.canonicalization',
 'dialogy.plugins.text.classification',
 'dialogy.plugins.text.combine_date_time',
 'dialogy.plugins.text.duckling_plugin',
 'dialogy.plugins.text.lb_plugin',
 'dialogy.plugins.text.list_entity_plugin',
 'dialogy.plugins.text.list_search_plugin',
 'dialogy.plugins.text.merge_asr_output',
 'dialogy.plugins.text.slot_filler',
 'dialogy.plugins.text.voting',
 'dialogy.types',
 'dialogy.types.entity',
 'dialogy.types.entity.amount_of_money',
 'dialogy.types.entity.base_entity',
 'dialogy.types.entity.credit_card_number',
 'dialogy.types.entity.deserialize',
 'dialogy.types.entity.duration',
 'dialogy.types.entity.keyword',
 'dialogy.types.entity.numerical',
 'dialogy.types.entity.people',
 'dialogy.types.entity.time',
 'dialogy.types.entity.time_interval',
 'dialogy.types.intent',
 'dialogy.types.plugin',
 'dialogy.types.signal',
 'dialogy.types.slots',
 'dialogy.types.utterances',
 'dialogy.utils',
 'dialogy.workflow']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=20.3.0,<21.0.0',
 'jiwer>=2.2.0,<3.0.0',
 'loguru>=0.5.3,<0.6.0',
 'numpy>=1.20.2,<2.0.0',
 'pandas>=1.3.2,<2.0.0',
 'pydash>=4.9.3,<5.0.0',
 'python-dateutil>=2.8.1,<3.0.0',
 'pytz>=2020.4,<2021.0',
 'requests>=2.25.1,<3.0.0',
 'scipy>=1.7.1,<2.0.0',
 'sklearn>=0.0,<0.1',
 'stanza>=1.3.0,<2.0.0',
 'thefuzz>=0.19.0,<0.20.0',
 'tqdm>=4.62.2,<5.0.0',
 'xgboost>=1.5.2,<2.0.0']

entry_points = \
{'console_scripts': ['dialogy = dialogy.cli:main']}

setup_kwargs = {
    'name': 'dialogy',
    'version': '0.9.4',
    'description': 'Dialogy is a library for building and managing SLU applications.',
    'long_description': None,
    'author': 'Amresh Venugopal',
    'author_email': 'amresh.venugopal@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/skit-ai/dialogy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)



"""
.. _BaseEntity:

Module provides access to a base class to create other entities.


.. mermaid::

    classDiagram
        direction LR
        BaseEntity --|> NumericalEntity
        BaseEntity --|> CreditCardNumberEntity
        BaseEntity --|> CurrencyEntity    
        BaseEntity --|> TimeEntity
        BaseEntity --|> KeywordEntity
        BaseEntity --|> DurationEntity
        BaseEntity --|> PeopleEntity

        TimeEntity --|> TimeIntervalEntity

        NumericalEntity ..> TimeEntity: as_time
        DurationEntity ..> TimeEntity: as_time

        class BaseEntity {
            +Dict[str, int] range
            +str body
            +str dim
            +int alternative_index
            +str type
            +str entity_type
            +float score
            +Any value
            +List~Any~ values
            +List~str~ parsers
            +bool latent
            +Any get_value()
            +Dict[str, Any] json()
            $BaseEntity from_dict(d)
        }

        class CurrencyEntity {
            +str unit
            +Any get_value()
            +CurrencyEntity from_duckling(d)
        }

        class CreditCardNumberEntity {
            +str issuer
            +from_duckling(d: dict, idx: int)
            +CreditCardNumberEntity from_duckling(d)
        }

        class KeywordEntity {
        }

        class DurationEntity {
            +str unit
            +dict normalized
            +DurationEntity from_duckling(d)
        }

        class NumericalEntity {
            +NumericalEntity from_duckling(d)
        }

        class PeopleEntity {
            +str unit
            +PeopleEntity from_duckling(d)
        }

        class TimeEntity {
            +str grain
            +datetime get_value()
            +TimeEntity from_duckling(d)
        }

"""
from __future__ import annotations

import copy
from typing import Any, Dict, List, Optional, Union

import attr

from dialogy import constants as const


# = BaseEntity =
@attr.s
class BaseEntity:
    """
    Base Entity Type.

    This class is meant for subclassing.
    Its intended purpose is to define a base for all entity types.
    """

    # **range**
    #
    # is the character range in the alternative where the entity is parsed.
    range: Dict[str, int] = attr.ib(repr=False, order=False, kw_only=True)

    # **body**
    #
    # is the string from which the entity is extracted.
    body: str = attr.ib(validator=attr.validators.instance_of(str), order=False)

    # **dim**
    #
    # is influenced from Duckling's convention of categorization.
    dim: Optional[str] = attr.ib(default=None, repr=False, order=False)

    # **type**
    #
    # is same as dimension or `dim` for now. We may deprecate `dim` and retain only `type`.
    type: str = attr.ib(default="value", order=False, kw_only=True)

    # **parsers**
    #
    # gives the trail of all the functions that have changed this entity in the sequence of application.
    parsers: List[str] = attr.ib(
        default=attr.Factory(list),
        validator=attr.validators.instance_of(list),
        order=False,
    )

    # **score**
    #
    # is the confidence that the range is the entity.
    score: Optional[float] = attr.ib(default=None)

    # **alternative_index**
    #
    # is the index of transcript within the ASR output: `List[Utterances]`
    # from which this entity was picked up. This may be None.
    alternative_index: Optional[int] = attr.ib(default=None, order=False)
    alternative_indices: Optional[List[int]] = attr.ib(default=None, order=False)

    # **latent**
    #
    # Duckling influenced attribute, tells if there is less evidence for an entity if latent is True.
    latent: bool = attr.ib(default=False, order=False)

    # **values**
    #
    # The parsed value of an entity resides within this attribute.
    values: List[Dict[str, Any]] = attr.ib(
        default=attr.Factory(list),
        validator=attr.validators.optional(attr.validators.instance_of(list)),  # type: ignore
        repr=False,
        order=False,
    )

    # **values**
    #
    # A single value interpretation from values.
    value: Any = attr.ib(default=None, order=False)

    # **entity_type**
    entity_type: Optional[str] = attr.ib(default=None, repr=True, order=False)

    def __attrs_post_init__(self) -> None:
        if self.values and not self.value:
            self.value = self.values[0][const.VALUE]
        elif self.value and not self.values:
            self.values = [{const.VALUE: self.value}]

    def add_parser(self, plugin: Union[Any, str]) -> "BaseEntity":
        """
        Update parsers with the postprocessor function name

        This is to identify the progression in which the plugins were applied to an entity.
        This only helps in debugging and has no production utility.

        :param plugin: The class that modifies this instance. Preferably should be a plugin.
        :type plugin: Plugin
        :return: Calling instance with modifications to :code:`parsers` attribute.
        :rtype: BaseEntity
        """
        self.parsers.append(str(plugin))
        return self

    def get_value(self) -> Any:
        """
        Get value of an entity.

        The structure of an entity conceals the value in different data structures.
        This is sugar for access.

        :param reference: Picking value/values from a reference object.
        :type reference: Any
        :raises IndexError: If `values` is not a `list`.
        :raises KeyError: If each element in `values` is not a `dict`.
        :return: Arbitrary instance, subclasses should override and return specific types.
        :rtype: BaseEntity
        """
        key = "value"
        error_message = f'entity value should be in values[0]["{key}"]'
        try:
            return self.values[0][key]
        except IndexError as index_error:
            raise IndexError(error_message) from index_error

    def copy(self) -> BaseEntity:
        """
        Create a deep copy of the instance and return.

        This is needed to prevent mutation of an original entity.
        :return: A copy of BaseEntity that calls this method.
        :rtype: BaseEntity
        """
        return copy.deepcopy(self)

    def json(
        self, skip: Optional[List[str]] = None, add: Optional[List[str]] = None
    ) -> Dict[str, Any]:
        """
        Convert the object to a dictionary.

        Applies some expected filters to prevent information bloat.

        Add section for skipping more properties using const.SKIP_ENTITY_ATTRS + [""]

        :param skip: Names of attributes that should not be included while converting to a dict.
            Defaults to None.
        :type skip: Optional[List[str]]
        :param add: Names of attributes that should be included while converting to a dict.
            Defaults to None.
        :type add: Optional[List[str]]
        :return: Dictionary representation of the object
        :rtype: Dict[str, Any]
        """
        skip_ = skip or const.SKIP_ENTITY_ATTRS
        if add and isinstance(add, list):
            skip_ = [name for name in const.SKIP_ENTITY_ATTRS if name not in add]
        return attr.asdict(self, filter=lambda attr, _: attr.name not in skip_)

    @classmethod
    def from_dict(
        cls, dict_: Dict[str, Any], reference: Optional[BaseEntity] = None
    ) -> BaseEntity:
        if reference:
            if const.VALUES in dict_ or const.VALUE in dict_:
                object.__setattr__(reference, const.VALUES, None)
                object.__setattr__(reference, const.VALUE, None)
            return attr.evolve(reference, **dict_)
        return cls(**dict_)


# = entity_synthesis =
def entity_synthesis(entity: BaseEntity, property_: str, value: Any) -> BaseEntity:
    """
    Update a property=`property_` of a BaseEntity, with a given value=`value`.

    .. warning:: This is an unsafe method. Use with caution as it doesn't check the type of
        the new value being type safe.

    :param entity: A BaseEntity instance.
    :type entity: BaseEntity
    :param property_: An attribute of BaseEntity that should be changed.
    :type property_: str
    :param value: The value to replace within a BaseEntity instance.
    :type value: Any
    :return: A copy of BaseEntity instance used in this function with modifications.
    :rtype: BaseEntity
    """
    if not isinstance(entity, BaseEntity):
        raise TypeError(f"{entity} has type {type(entity)} expected BaseEntity")
    synthetic_entity = entity.copy()
    setattr(synthetic_entity, property_, value)
    return synthetic_entity



import re
from typing import List, Union


def _lang_detect(text: str) -> str:
    text = re.sub(r"[^a-z\u0900-\u0c7f]+", " ", text.lower())
    maxchar = max(text)
    if "\u0c00" <= maxchar <= "\u0c7f":
        return "te"
    if "\u0900" <= maxchar <= "\u097f":
        return "hi"
    return "en"


def lang_detect_from_text(text: Union[str, List[str]]) -> str:
    if isinstance(text, str):
        return _lang_detect(text)
    if text and isinstance(text, list):
        return _lang_detect(" ".join(text))
    raise TypeError(f"{text=} must be a string or a list of strings")



"""
We use classifiers for prediction of :ref:`intents<Intent>`, utterances from an ASR are prominent features
for the task. We featurize the utterances by concatenation of all transcripts.

.. ipython::

    In [1]: from dialogy.workflow import Workflow
       ...: from dialogy.plugins import MergeASROutputPlugin
       ...: from dialogy.base import Input

    In [2]: merge_asr_output_plugin = MergeASROutputPlugin(dest="input.clf_feature")
       ...: workflow = Workflow([merge_asr_output_plugin])

    In [3]: input_, _ = workflow.run(Input(utterances=["we will come by 7 pm", " will come by 7 pm"]))

    In [4]: input_
"""
import json
import traceback
from typing import Any, List, Optional

import pandas as pd
from loguru import logger
from tqdm import tqdm

import dialogy.constants as const
from dialogy.base import Guard, Input, Output, Plugin
from dialogy.utils import normalize


# == merge_asr_output ==
def merge_asr_output(utterances: Any) -> List[str]:
    """
    .. _merge_asr_output:

    Join ASR output to single string.

    This function provides a merging strategy for n-best ASR transcripts by
    joining each transcript, such that:

    - each sentence end is marked by " </s>" and,
    - sentence start marked by " <s>".

    The key "transcript" is expected in the ASR output, the value of which would be operated on
    by this function.

    The normalization is done by :ref:`normalize<normalize>`

    :param utterances: A structure representing ASR output. We support only:

        1. :code:`List[str]`
        2. :code:`List[List[str]]`
        3. :code:`List[List[Dict[str, str]]]`
        4. :code:`List[Dict[str, str]]`

    :type utterances: Any
    :return: Concatenated string, separated by <s> and </s> at respective terminal positions of each sentence.
    :rtype: List[str]
    :raises: TypeError if transcript is missing in cases of :code:`List[List[Dict[str, str]]]` or
        :code:`List[Dict[str, str]]`.
    """
    try:
        transcripts: List[str] = normalize(utterances)
        invalid_transcript = len(transcripts) == 1 and any(
            token.lower() in transcripts for token in const.INVALID_TOKENS
        )
        if invalid_transcript or not transcripts:
            return []
        else:
            return ["<s> " + " </s> <s> ".join(transcripts) + " </s>"]
    except TypeError as type_error:
        raise TypeError("`transcript` is expected in the ASR output.") from type_error


class MergeASROutputPlugin(Plugin):
    """
    .. _MergeASROutputPlugin:

    Working details are covered in :ref:`MergeASROutputPlugin<MergeASROutputPlugin>`.

    :param Plugin: [description]
    :type Plugin: [type]
    """

    def __init__(
        self,
        input_column: str = const.ALTERNATIVES,
        output_column: Optional[str] = None,
        use_transform: bool = False,
        dest: Optional[str] = None,
        guards: Optional[List[Guard]] = None,
        debug: bool = False,
    ) -> None:
        super().__init__(
            dest=dest,
            guards=guards,
            input_column=input_column,
            output_column=output_column,
            use_transform=use_transform,
            debug=debug,
        )

    def utility(self, input: Input, _: Output) -> Any:
        return merge_asr_output(input.utterances)

    def transform(self, training_data: pd.DataFrame) -> pd.DataFrame:
        if not self.use_transform:
            return training_data

        training_data["use"] = True
        logger.debug(f"Transforming dataset via {self.__class__.__name__}")
        for i, row in tqdm(training_data.iterrows(), total=len(training_data)):
            asr_output = None
            try:
                asr_output = json.loads(row[self.input_column])
                if asr_output and (merged_asr_ouptut := merge_asr_output(asr_output)):
                    training_data.loc[i, self.output_column] = merged_asr_ouptut[0]
                else:
                    training_data.loc[i, "use"] = False
            except Exception as error:  # pylint: disable=broad-except
                training_data.loc[i, "use"] = False
                logger.error(f"{error} -- {asr_output}\n{traceback.format_exc()}")

        training_data_ = training_data[training_data.use].copy()
        training_data_.drop("use", axis=1, inplace=True)
        discarded_data = len(training_data) - len(training_data_)
        if discarded_data:
            logger.debug(
                f"Discarding {discarded_data} samples because the alternatives couldn't be parsed."
            )
        return training_data_



"""Module provides Constants to be used throughtout the project."""

from locale import DAY_1

DEFAULT_NAMESPACE = "vernacular-ai"
DEFAULT_PROJECT_TEMPLATE = "dialogy-template-simple-transformers"

PLUGINS = "plugins"
DEBUG = "debug"
S_INTENT_OOS = "_oos_"
TRANSCRIPT = "transcript"
AM_SCORE = "am_score"
LM_SCORE = "lm_score"
VALUE = "value"
TO = "to"
FROM = "from"
LOCK = "lock"
NAME = "name"
SLOTS = "slots"

DUCKLING_ENTITY_KEYS = ("body", "dim", "end", "latent", "start", "value")
DUCKLING_TIME_VALUES_ENTITY_KEYS = ("grain", "type", "value", "values")
DUCKLING_TIME_INTERVAL_ENTITY_KEYS = (
    ("from", "type", "values"),
    ("to", "type", "values"),
    ("from", "to", "type", "values"),
)

# Time Entity types
DATE = "date"
TIME = "time"
DATETIME = "datetime"
HOUR = "hour"
DAY = "day"
MONTH = "month"

PEOPLE = "people"
NUMBER = "number"
AMOUNT_OF_MONEY = "amount-of-money"
DURATION = "duration"
CREDIT_CARD_NUMBER = "credit-card-number"

FUTURE = "future"
PAST = "past"
ANY = "__any__"

DUCKLING_DIMS = [PEOPLE, NUMBER, AMOUNT_OF_MONEY, DURATION, CREDIT_CARD_NUMBER, TIME]

BODY = "body"
DIM = "dim"
END = "end"
FROM = "from"
GRAIN = "grain"
INTERVAL = "interval"
DURATION = "duration"
LATENT = "latent"
RANGE = "range"
START = "start"
TO = "to"
TYPE = "type"
ENTITY_TYPE = "entity_type"
UNIT = "unit"
VALUE = "value"
VALUES = "values"
ORIGIN = "origin"
ISSUER = "issuer"
NORMALIZED = "normalized"
ALTERNATIVE_INDEX = "alternative_index"
ALTERNATIVE_INDICES = "alternative_indices"
CREDIT_CARD_NUMBER = "credit-card-number"

SLOT_NAMES = "slot_names"
SKIP_ENTITY_ATTRS = [
    DIM,
    VALUES,
    LATENT,
    ORIGIN,
    SLOT_NAMES,
    ALTERNATIVE_INDICES,
]

# This section is needed for dialogy.types.entities.*
# Each entity type has a few properties which need to be
# validated for their type.
#
# The actual use-case is, a `dict` is received by the factory
# method of the Entity class, which provides all/enough values to
# instantiate the Entity. We use this mapping to ensure that the
# values belong to expected types.
BASE_ENTITY_PROPS = [
    ([RANGE], dict),
    ([RANGE, START], int),
    ([RANGE, END], int),
    ([BODY], str),
    ([VALUES], list),
    ([DIM], str),
    ([LATENT], bool),
]

CREDIT_CARD_NUMBER_PROPS = [
    ([RANGE], dict),
    ([RANGE, START], int),
    ([RANGE, END], int),
    ([BODY], str),
    ([VALUE], dict),
    ([DIM], str),
    ([LATENT], bool),
]

TIME_ENTITY_PROPS = BASE_ENTITY_PROPS + [([GRAIN], str)]
PEOPLE_ENTITY_PROPS = BASE_ENTITY_PROPS + [([UNIT], str)]

# Time related constants
DATE_UNITS = ["day", "week", "month", "quarter", "year"]
TIME_UNITS = ["hour", "minute", "second"]


# This section covers signals tuple indices
class SIGNAL:
    NAME = 0
    STRENGTH = 1
    REPRESENTATION = 2


KEYWORD = "keyword"
REGEX = "regex"
SPACY = "spacy"
PWER = "pWER"
ALTERNATIVES = "alternatives"
UTTERANCES = "utterances"

# Transcripts containing less than or equal to this number of for words.
WORD_THRESHOLD = 2

# XLMR
ID = "id"
IDX = "idx"
DATA = "data"
STATE = "state"
CLASSIFICATION_INPUT = "classification_input"
XLMR_MODULE = "simpletransformers.classification"
XLMR_MULTI_CLASS_MODEL = "ClassificationModel"
XLMR_MODEL = "xlmroberta"
XLMR_MODEL_TIER = "xlm-roberta-base"
INTENT = "intent"
INTENTS = f"{INTENT}s"
LABELS = "labels"
ENTITIES = "entities"
SCORE = "score"
LABELENCODER_FILE = "labelencoder.pkl"
INVALID_TOKENS = ["<UNK>", "<PAD>", "<SOS>", "<EOS>"]

# MLP
MLPMODEL_FILE = "mlpmodelpipeline.joblib"
NUM_TRAIN_EPOCHS = "num_train_epochs"
MLP_DEFAULT_TRAIN_EPOCHS = 25
DEFAULT_NGRAM = (1, 1)
TFIDF_LOWERCASE = False
STOPWORDS = None
MLP_RANDOMSTATE = 1

TFIDF = "tfidf"
MLP = "mlp"

# Hyperparams For MLP
USE_GRIDSEARCH = "gridsearch_hyperparams"
PARAMS = "params"
GRID_SCORETYPE = "weighted"
CV = "cv"
VERBOSE_LEVEL = "verbose_level"
NGRAM_RANGE = "ngram_range"
GRIDSEARCH_WORKERS = -1  # -1 means use all available cores

# CLI Commands
TRAIN = "train"
TEST = "test"
PRODUCTION = "production"
CREATE = "create"
UPDATE = "update"
ERROR_LABEL = "_error_"
TEXT = "text"
CANONICALIZED_TRANSCRIPTS = "canonicalized_transcripts"
REFERENCE_TIME = "reftime"
ENTITY_COLUMN = "entities"

FUZZY_DP = "fuzzy_dp"



from typing import Any, Dict, List, Optional

from pydash import partition

from dialogy import constants as const
from dialogy.base import Guard, Input, Output
from dialogy.plugins import DucklingPlugin
from dialogy.types import BaseEntity


class DucklingPluginLB(DucklingPlugin):
    # Constructor
    def __init__(
        self,
        dimensions: List[str],
        timezone: str,
        timeout: float = 0.5,
        url: str = "http://0.0.0.0:8000/parse",
        locale: str = "en_IN",
        datetime_filters: Optional[str] = None,
        dest: Optional[str] = None,
        guards: Optional[List[Guard]] = None,
        reference_time_column: str = const.REFERENCE_TIME,
        input_column: str = const.ALTERNATIVES,
        output_column: Optional[str] = None,
        use_transform: bool = False,
        debug: bool = False,
    ):
        super().__init__(
            dimensions,
            timezone,
            timeout=timeout,
            url=url,
            locale=locale,
            datetime_filters=datetime_filters,
            threshold=0,
            dest=dest,
            guards=guards,
            reference_time_column=reference_time_column,
            input_column=input_column,
            output_column=output_column,
            use_transform=use_transform,
            debug=debug,
        )

    def utility(self, input_: Input, output: Output) -> List[BaseEntity]:
        entity_list = super().utility(input_, output)
        datetime_list, other_list = partition(
            entity_list, lambda x: x.entity_type in ["datetime", "date", "time"]
        )
        if datetime_list:
            other_list.append(min(datetime_list, key=lambda x: x.alternative_index))

        return other_list
